# -*- coding: UTF-8 -*-

import json
import os
import sys
import threading
import time
import traceback
from functools import wraps
from bottle import request, response, HTTPResponse, Bottle
from common import logger, config, filesmodify, parameters
from common.rest_state import RestState as rest_state
from upload.fam.session.restSessionService import RestSessionService
from upload.node.node import DBInserterConfig, QueueInserterConfig
if sys.version_info.major == 3:
    from http.client import INTERNAL_SERVER_ERROR, UNAUTHORIZED
else:
    from httplib import INTERNAL_SERVER_ERROR, UNAUTHORIZED

class RestService(Bottle):

    def __init__(self, isCash=True):
        super(RestService, self).__init__()
        self.__readConfig()
        self.log = logger.getUploadLogger('rest')
        self.restLogName = 'dictLoadRestBackup'
        self.isCash = isCash
        self.restServiceHost = '0.0.0.0'
        self.route('/aifexchange/loaddicts', method='POST', callback=self.__restStatus)
        self.route('/aifexchange/loaddicts/session/start', method='POST', callback=self.__start_session)
        self.route('/aifexchange/loaddicts/session/stop', method='POST', callback=self.__stop_session)
        self.route('/aifexchange/loaddicts/session/rollback', method='POST', callback=self.__rollback_session)
        self.sessionService = RestSessionService(isCash, self.defaultNesServer, self.defaultStatusServer)
        self.sessionService.useSession = self.useSession
        self.sessionService.singleShopSession = self.singleShopSession
        self.sessionService.sessionIdleTimeout = self.sessionIdleTimeout
        self.lastTimeReadNodes = time.time() - self.timeoutNodeReread
        self.nodes = []

    def start(self):
        self.log.info('Запуск REST-сервиса загрузки справочников, host = %s, port = %s' % (self.restServiceHost, self.restServicePort))
        if self.useSession:
            self.log.info("Starting thread for idle session")
            thread = threading.Thread(target=self.session_thread_function)
            thread.daemon = True
            thread.start()

        try:
            self.install(self.__loggingRequest)
            self.run(host=self.restServiceHost, port=self.restServicePort, quiet=True)
        except Exception as e:
            self.log.error(u'Ошибка: %s' % e)
            self.log.error(traceback.format_exc())

    def __readConfig(self):
        self.primitiveAuth = config.getBool('REST_UPLOAD', 'authorize', True)
        self.restLogin = config.getString('REST_UPLOAD', 'login', 'admin')
        self.restPasswd = config.getString('REST_UPLOAD', 'passwd', 'admin')
        self.restServicePort = config.getInt('REST_UPLOAD', 'port', 28080)
        self.shopCodeIsShopId = config.getBool('REST_UPLOAD', 'shopCodeIsShopId', True)
        self.maxCountBackupRequests = config.getInt('REST_UPLOAD', 'maxCountBackupRequests', 0)
        self.actualCountDaysForBackupRequests = config.getInt('REST_UPLOAD', 'actualCountDaysForBackupRequests', 3)
        self.defaultStatusServer = config.getString('CS', 'defaultStatusServer', 'localhost:8080')
        self.defaultNesServer = config.getString('CS', 'defaultNesServer', 'localhost:28081')
        self.timeoutNodeReread = config.getInt('UPLOAD', 'timeoutNodeReread', 300)
        self.shopCodeIsShopId = config.getBool('REST_UPLOAD', 'shopCodeIsShopId', True)
        self.sessionIdleTimeout = config.getInt('REST_UPLOAD', 'sessionIdleTimeout', 300)
        self.useSession = config.getBool('REST_UPLOAD', 'useSession', False)
        self.singleShopSession = config.getBool('REST_UPLOAD', 'singleShopSession', False)

    def __authorization(func):
        """Проверка авторизации"""
        def check_auth(self):
            if self.primitiveAuth:
                if request.auth and request.auth == (self.restLogin, self.restPasswd):
                    return func(self)
                else:
                    raise HTTPResponse("Unauthorized", status=UNAUTHORIZED)
            return func(self)
        return check_auth

    @__authorization
    def __restStatus(self):
        """Обработка входящего запроса aifexchange/loaddicts с провекой авторизации"""
        response.content_type = 'application/json'
        body = None
        dictionaryId = request.query.get('dictionaryid', None)
        node = None
        try:
            if request.content_length > 0 or request.chunked:
                body = json.loads(request.body.read(), object_hook=self.__filterKeys)
            if not isinstance(body, list):
                raise Exception('Body is not array.')
            if len(body) == 0:
                body = None
                raise Exception('Array is empty.')
            node, result = self.__load(body, shopCode=request.query.get('shopcode', 'All'),
                                       dictionaryId=dictionaryId, statusServer=request.query.get('addressstatusserver', None))

        except Exception as e:
            errorStr = u''
            errorStr_ru = u''
            if isinstance(body, list):
                errorStr = u'Error loading dicts. %s' % str(e)
                errorStr_ru = u'Ошибка загрузки справочника. %s' % str(e)
            else:
                errorStr = u'Incorrect request. %s' % str(e)
                errorStr_ru = u'Неправильный запрос. %s' % str(e)
            self.log.error(errorStr)
            self.log.error(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
            result = {u'success': False, u'info': errorStr, u'info_ru': errorStr_ru}
        if node:
            rest_state.sendDictState(node, result['info'], result['info_ru'], rest_state.STATUS_READY if result['success'] else rest_state.STATUS_ERROR,
                                     dictionaryId)

        return json.dumps(result, ensure_ascii=False)

    @__authorization
    def __start_session(self):
        status_server = request.query.get('addressstatusserver', None)
        shop_code = request.query.get('shopcode')
        dictionary_id = request.query.get('dictionaryid')
        result, shop_id = self.__checkShopCode(shop_code)
        node = EmptyNode(self.isCash, shop_code, shop_id, status_server if status_server else self.defaultStatusServer, dictionary_id, self.defaultNesServer)
        if result:
            return json.dumps(self.sessionService.start_session(dictionary_id, node), ensure_ascii=False)
        else:
            result = None, {u'success': False, 'info': "Shop with code %s not found" % shop_code, 'info_ru': u'Магазин с кодом = %s не найден.' % shop_code}
            return json.dumps(result, ensure_ascii=False)

    @__authorization
    def __stop_session(self):
        shop_code = request.query.get('shopcode')
        result, shop_id = self.__checkShopCode(shop_code)
        if result:
            return json.dumps(self.sessionService.stop_session(shop_code, request.query.get('dictionaryid'), shop_id), ensure_ascii=False)
        else:
            result = None, {u'success': False, 'info': "Shop with code %s not found" % shop_code, 'info_ru': u'Магазин с кодом = %s не найден.' % shop_code}
            return json.dumps(result, ensure_ascii=False)

    @__authorization
    def __rollback_session(self):
        shop_code = request.query.get('shopcode')
        result, shop_id = self.__checkShopCode(shop_code)
        if result:
            return json.dumps(self.sessionService.rollback_session(shop_code, request.query.get('dictionaryid'), shop_id=shop_id), ensure_ascii=False)
        else:
            result = None, {u'success': False, 'info': "Shop with code %s not found" % shop_code, 'info_ru': u'Магазин с кодом = %s не найден.' % shop_code}
            return json.dumps(result, ensure_ascii=False)

    def session_thread_function(self):
        try:
            while True:
                sessions = self.sessionService.sessions
                for k in list(sessions):
                    if (time.time() - sessions[k].timeActive) > self.sessionIdleTimeout:
                        self.sessionService.rollback_session(shop_code=sessions[k].node.shopCode, shop_id=sessions[k].node.id, dictionary_id=sessions[k].dictId, session_idle=True)
                time.sleep(2)
        except Exception as e:
            self.log.error(e)

    @staticmethod
    def __filterKeys(d):
        """используется в object_hook - Optional - функции, скорее всего, чтобы избежать None"""
        lowerD = {}
        for k in d.keys():
            if d[k] is not None and d[k] != []:
                lowerD[k.lower()] = d[k]
        return lowerD

    def __load(self, dictObjects, shopCode='All', dictionaryId=None, statusServer=None):
        """Загрузка справочника"""
        self.log.info("Загрузка через очередь")
        self.__readNodes()


        realShopCode = shopCode
        result, shop_id = self.__checkShopCode(shopCode)
        if not result:
            return None, {u'success': False, 'info': "Shop with code %s not found" % shopCode, 'info_ru': u'Магазин с кодом = %s не найден.' % shopCode}
        self.log.info('status server from request=%s' % statusServer)
        if self.useSession:
            if not self.sessionService.is_open_session(shopCode, dictionaryId):
                info = "Can not add package. Session for shopCode: %s and dictionaryId: %s is not opened" % (realShopCode, dictionaryId)
                self.log.warning(info)
                return None, {u'success': False, 'info': info,
                              'info_ru': u'Не удалось добавить справочник. Сессия для магазина с кодом: %s и справочника: %s не открыта' % (shopCode, dictionaryId)}
            self.log.info('Add package in session %s for dictionaryId %s' %
                          (self.sessionService.sessions[self.sessionService.get_session_id(shop_id,
                                                                                           dictionaryId)].node.currentConverter.dict_state_id,
                           dictionaryId))
            result = True
            self.sessionService.sessions[
                self.sessionService.get_session_id(shop_id, dictionaryId)].timeActive = time.time()
            node = self.sessionService.sessions[self.sessionService.get_session_id(shop_id, dictionaryId)].node
            inserter = self.sessionService.sessions[self.sessionService.get_session_id(shop_id, dictionaryId)].inserter
        else:
            node = EmptyNode(self.isCash, shopCode, shop_id, statusServer if statusServer else self.defaultStatusServer, dictionaryId,
                             self.defaultNesServer)
            node.setNodeConverter(self.nodes)
            plugin = __import__('upload.inserter.%s' % node.inserter.name, fromlist='Inserter')
            cl = getattr(plugin, 'Inserter')
            inserter = cl(node)
            inserter.disableProgressFile = False
            inserter.source = "exchangers-rest-upload"
            try:
                is_inserter_start = inserter.start()
                if not is_inserter_start:
                    return node, {u'success': is_inserter_start, 'info': u'Can not start for load to queue.',
                                  'info_ru': u'Невозможно начать загрузку в очередь.'}
            except Exception as e:
                self.log.error(u'Ошибка: {0}'.format(unicode(e)))
                self.log.error(traceback.format_exc())
                return node, {u'success': False, 'info': u'Error: ' + unicode(e), 'info_ru': u'Ошибка: ' + unicode(e)}

        if node.currentConverter.hotKeyByBarcode:
            for item in dictObjects:
                if 'addHotKey' == item['command'] and not 'bybarcode' in item['hotkey']:
                    item['hotkey']['bybarcode'] = node.currentConverter.hotKeyByBarcode

        n = len(dictObjects)
        k = n // node.pack_size
        if n % node.pack_size > 0:
            k += 1
        for i in range(k):
            inserter.add_package(dictObjects[i * node.pack_size:(i * node.pack_size) + node.pack_size])

        if not self.useSession:
            result = inserter.stop()
            if result:
                return node, {u'success': result, 'info': u'OK', u'info_ru': 'OK'}
            else:
                return node, {u'success': result, 'info': u'Can not commit to queue.',
                              'info_ru': u'Невозможно положить в очередь объекты'}
        return node, {u'success': result, 'info': u'Dict added to session load',
                      'info_ru': u'Справочник добавлен в сессионную загрузку'}

    def __checkShopCode(self, shopCode):
        if self.isCash:
            return True, shopCode
        if not shopCode or shopCode.strip() == '':
            return False, shopCode
        if self.nodes:
            if not self.shopCodeIsShopId:
                ids = [n.id for n in self.nodes if str(n.shopCode) == shopCode]
                if len(ids) == 1:
                    shopCode = ids[0]
                else:
                    self.log.warning(u'По коду магазина %s найдено несколько идентификаторов либо ни одного. Невозможно однозначно определить целевой магазин.' % shopCode)
                    return False, shopCode
            elif not str(shopCode) in [str(node.id) for node in self.nodes]:
                return False, shopCode
        elif not self.shopCodeIsShopId:
            self.log.warning(u'Невозможно найти по коду магазина его идентификатор, т.к. ноды не считаны')
            return False, shopCode
        return True, shopCode

    def __readNodes(self):
        """чтение нод на кассе/кс"""

        nodeGetter = __import__('upload.node.%s.nodeGetter' % (config.getString('UPLOAD', 'nodeGetter', 'cash' if self.isCash else 'cs30')),
                                fromlist='q')
        nodes = []
        if time.time() - self.lastTimeReadNodes > self.timeoutNodeReread or not self.nodes:
            self.lastTimeReadNodes = time.time()
            try:
                nodeGetterLog = nodeGetter.log
                nodeGetterLog.info(u'Чтение нод')
                nodeGetter.defaultNesServer = self.defaultNesServer
                nodeGetter.defaultStatusServer = self.defaultStatusServer
                nodes = nodeGetter.getNodes()
                nodeGetterLog.debug(u'ноды = %s' % ' '.join(str(e) for e in nodes))
            except Exception as e:
                nodeGetterLog.warning(u'Ошибка считывания нод. %s' % str(e))
                nodeGetterLog.warning(traceback.format_exc())
        # записываем полученные ноды в глобальный кэш, т.о. в случае ошибок считывания узлов с КС будем брать их из кэша
        if nodes:
            self.nodes = nodes


    def __loggingRequest(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            """Логирование входящего запроса"""
            self.log.info('Получен запрос от %s' % request.remote_addr)
            backuplog = logger.getUploadLogger('RestBackup')
            backuplog.debug('Получен запрос от %s' % request.remote_addr)
            backuplog.debug('remote_addr = %s ' % request.remote_addr)
            backuplog.debug('url = %s ' % request.url)
            backuplog.debug('method = %s ' % request.method)
            backuplog.debug('headers = %s ' % (request.headers.items()))
            backuplog.debug('content_length = %s ' % request.content_length)
            backuplog.debug('chunked = %s ' % request.chunked)
            if request.content_length > 0 or request.chunked:
                backuplog.debug('body = %s ' % request.body.read())
            self.log.info('Получен запрос от %s' % request.remote_addr)
            try:
                actual_response = fn(*args, **kwargs)
            except Exception as e:
                self.log.error(traceback.format_exc())
                raise e
            self.log.info('Обработка запроса завершена')
            self.log.info(actual_response)
            backuplog.debug(actual_response)
            backuplog.debug('Обработка запроса завершена успешно')
            self.__backupRequest(request.query.get('shopcode', 'All'))
            return actual_response

        return wrapper

    def __backupRequest(self, shopCode):
        """Архивирование лога с полученным запросом во избежание разрастания логов"""
        logDir = logger.getLogDirectory()
        logFiles = filesmodify.getFilesByMask(logDir, ['%s.*' % self.restLogName])
        for f in logFiles:
            filesmodify.archive(self.log, os.path.dirname(f), os.path.join(logDir, 'REST_BACKUP', shopCode), maxCountFiles=self.maxCountBackupRequests,
                                actualCountDays=self.actualCountDaysForBackupRequests, removeSRC=True, fileName=os.path.basename(f), loginfo=False)


class EmptyNode(object):

    def __init__(self, isCashRun, shopCode, shop_id, statusServer=None, dictionaryId=None, nesServer = None):
        self.shopCode = shopCode
        self.id = shop_id
        self.dictionaryId = dictionaryId
        self.extendedProperties = {}
        self.cashCode = ""
        self.nesServer = nesServer
        self.statusServer = statusServer
        self.currentConverter = self.RestConverter()
        self.pack_size = config.getInt('FILEQUEUE', 'queueChunkSize', 25000)
        self.queuepath = config.getString('FILEQUEUE', 'fileQueueRoot', '/linuxcash/cash/data/queues' if isCashRun else '/opt/queues')
        self.inserter = QueueInserterConfig(
            queue=parameters.QueueParameters(self.queuepath, self.pack_size, 'artix.loaddict.cash' if isCashRun else None))

        self.log = logger.getUploadLogger("Node", self.id)
        self.dictionaryId = 'rest' + str(time.time()) if not self.dictionaryId else self.dictionaryId
        # в ноде уже есть статус сервер, но есть вариант когда статус сервер передается в запросе, и ответы необходимо слать на него (Губернский лекарь,
        # загрузска справочников через REST на прямую в кассу)
        self.currentConverter.statusServer = self.statusServer
        try:
            self.currentConverter.dict_state_id = rest_state.sendDictState(self, 'Start convert dict', u'Начало конвертирования справочника.',
                                                                           rest_state.STATUS_DURING, self.dictionaryId)
            if not self.currentConverter.dict_state_id:
                self.currentConverter.dict_state_id = self.dictionaryId
        except Exception as e:
            self.log.warning(u'Ошибка отправки статуса на КС через REST. %s' % e)
            self.log.warning(traceback.format_exc())

    def setNodeConverter(self, nodes):
        if nodes:
            for n in nodes:
                if str(n.id) == str(self.id):
                    self.inserter = n.inserter
                    for converter in n.converters:
                        if converter.name == 'aif':
                            self.currentConverter.hotKeyByBarcode = converter.hotKeyByBarcode
                    break

    class RestConverter(object):
        """Конвертер для ноды"""
        name = 'rest'
        dict_state_id = None

        def __init__(self, hotKeyByBarcode=False):
            self.hotKeyByBarcode = hotKeyByBarcode
